@@ -72,6 +72,8 @@ int Manage_ro(char *devname, int fd, int readonly)
 	return 0;			
 }
 
+#ifndef MDASSEMBLE
+
 int Manage_runstop(char *devname, int fd, int runstop, int quiet)
 {
 	/* Run or stop the array. array must already be configured
@@ -104,7 +106,11 @@ int Manage_runstop(char *devname, int fd, int runstop, int quiet)
 				devname, strerror(errno));
 			return 1;
 		}
+		if (quiet <= 0)
+			fprintf(stderr, Name ": started %s\n", devname);
 	} else if (runstop < 0){
+		struct map_ent *map = NULL;
+		struct stat stb;
 		if (ioctl(fd, STOP_ARRAY, NULL)) {
 			if (quiet==0)
 				fprintf(stderr, Name ": fail to stop array %s: %s\n",
@@ -113,6 +119,16 @@ int Manage_runstop(char *devname, int fd, int runstop, int quiet)
 		}
 		if (quiet <= 0)
 			fprintf(stderr, Name ": stopped %s\n", devname);
+		if (fstat(fd, &stb) == 0) {
+			int devnum;
+			if (major(stb.st_rdev) == MD_MAJOR)
+				devnum = minor(stb.st_rdev);
+			else
+				devnum = -1-(minor(stb.st_rdev)>>6);
+			map_delete(&map, devnum);
+			map_write(map);
+			map_free(map);
+		}
 	}
 	return 0;
 }
@@ -164,13 +180,17 @@ int Manage_subdevs(char *devname, int fd,
 	 *	   try HOT_ADD_DISK
 	 *         If that fails EINVAL, try ADD_NEW_DISK
 	 *  'r' - remove the device HOT_REMOVE_DISK
+	 *        device can be 'faulty' or 'detached' in which case all
+	 *	  matching devices are removed.
 	 *  'f' - set the device faulty SET_DISK_FAULTY
+	 *        device can be 'detached' in which case any device that
+	 *	  is inaccessible will be marked faulty.
 	 */
 	mdu_array_info_t array;
 	mdu_disk_info_t disc;
-	mddev_dev_t dv;
+	mddev_dev_t dv, next = NULL;
 	struct stat stb;
-	int j;
+	int j, jnext = 0;
 	int tfd;
 	struct supertype *st;
 	void *dsuper = NULL;
@@ -183,16 +203,86 @@ int Manage_subdevs(char *devname, int fd,
 			devname);
 		return 1;
 	}
-	for (dv = devlist ; dv; dv=dv->next) {
-		if (stat(dv->devname, &stb)) {
-			fprintf(stderr, Name ": cannot find %s: %s\n",
-				dv->devname, strerror(errno));
-			return 1;
-		}
-		if ((stb.st_mode & S_IFMT) != S_IFBLK) {
-			fprintf(stderr, Name ": %s is not a block device.\n",
-				dv->devname);
-			return 1;
+	for (dv = devlist, j=0 ; dv; dv = next, j = jnext) {
+		unsigned long long ldsize;
+		char dvname[20];
+		char *dnprintable = dv->devname;
+
+		next = dv->next;
+		jnext = 0;
+
+		if (strcmp(dv->devname, "failed")==0 ||
+		    strcmp(dv->devname, "faulty")==0) {
+			if (dv->disposition != 'r') {
+				fprintf(stderr, Name ": %s only meaningful "
+					"with -r, not -%c\n",
+					dv->devname, dv->disposition);
+				return 1;
+			}
+			for (; j < array.raid_disks + array.nr_disks ; j++) {
+				disc.number = j;
+				if (ioctl(fd, GET_DISK_INFO, &disc))
+					continue;
+				if (disc.major == 0 && disc.minor == 0)
+					continue;
+				if ((disc.state & 1) == 0) /* faulty */
+					continue;
+				stb.st_rdev = makedev(disc.major, disc.minor);
+				next = dv;
+				jnext = j+1;
+				sprintf(dvname,"%d:%d", disc.major, disc.minor);
+				dnprintable = dvname;
+				break;
+			}
+			if (jnext == 0)
+				continue;
+		} else if (strcmp(dv->devname, "detached") == 0) {
+			if (dv->disposition != 'r' && dv->disposition != 'f') {
+				fprintf(stderr, Name ": %s only meaningful "
+					"with -r of -f, not -%c\n",
+					dv->devname, dv->disposition);
+				return 1;
+			}
+			for (; j < array.raid_disks + array.nr_disks; j++) {
+				int sfd;
+				disc.number = j;
+				if (ioctl(fd, GET_DISK_INFO, &disc))
+					continue;
+				if (disc.major == 0 && disc.minor == 0)
+					continue;
+				sprintf(dvname,"%d:%d", disc.major, disc.minor);
+				sfd = dev_open(dvname, O_RDONLY);
+				if (sfd >= 0) {
+					close(sfd);
+					continue;
+				}
+				if (dv->disposition == 'f' &&
+				    (disc.state & 1) == 1) /* already faulty */
+					continue;
+				if (errno != ENXIO)
+					continue;
+				stb.st_rdev = makedev(disc.major, disc.minor);
+				next = dv;
+				jnext = j+1;
+				dnprintable = dvname;
+				break;
+			}
+			if (jnext == 0)
+				continue;
+		} else {
+			j = 0;
+
+			if (stat(dv->devname, &stb)) {
+				fprintf(stderr, Name ": cannot find %s: %s\n",
+					dv->devname, strerror(errno));
+				return 1;
+			}
+			if ((stb.st_mode & S_IFMT) != S_IFBLK) {
+				fprintf(stderr, Name ": %s is not a "
+					"block device.\n",
+					dv->devname);
+				return 1;
+			}
 		}
 		switch(dv->disposition){
 		default:
@@ -200,7 +290,7 @@ int Manage_subdevs(char *devname, int fd,
 				dv->devname, dv->disposition);
 			return 1;
 		case 'a':
-			/* add the device - hot or cold */
+			/* add the device */
 			st = super_by_version(array.major_version,
 					      array.minor_version);
 			if (!st) {
@@ -219,6 +309,10 @@ int Manage_subdevs(char *devname, int fd,
 			if (array.not_persistent==0)
 				st->ss->load_super(st, tfd, &osuper, NULL);
 			/* will use osuper later */
+			if (!get_dev_size(tfd, dv->devname, &ldsize)) {
+				close(tfd);
+				return 1;
+			}
 			close(tfd);
 
 			if (array.major_version == 0 &&
@@ -238,6 +332,14 @@ int Manage_subdevs(char *devname, int fd,
 
 			if (array.not_persistent == 0) {
 
+				/* Make sure device is large enough */
+				if (st->ss->avail_size(st, ldsize/512) <
+				    array.size) {
+					fprintf(stderr, Name ": %s not large enough to join array\n",
+						dv->devname);
+					return 1;
+				}
+
 				/* need to find a sample superblock to copy, and
 				 * a spare slot to use 
 				 */
@@ -259,6 +361,7 @@ int Manage_subdevs(char *devname, int fd,
 						close(dfd);
 						continue;
 					}
+					remove_partitions(dfd);
 					close(dfd);
 					break;
 				}
@@ -300,6 +403,15 @@ int Manage_subdevs(char *devname, int fd,
 						/* fall back on normal-add */
 					}
 				}
+			} else {
+				/* non-persistent. Must ensure that new drive
+				 * is at least array.size big.
+				 */
+				if (ldsize/512 < array.size) {
+					fprintf(stderr, Name ": %s not large enough to join array\n",
+						dv->devname);
+					return 1;
+				}
 			}
 			/* in 2.6.17 and earlier, version-1 superblocks won't
 			 * use the number we write, but will choose a free number.
@@ -369,27 +481,43 @@ int Manage_subdevs(char *devname, int fd,
 			/* hot remove */
 			/* FIXME check that it is a current member */
 			if (ioctl(fd, HOT_REMOVE_DISK, (unsigned long)stb.st_rdev)) {
-				fprintf(stderr, Name ": hot remove failed for %s: %s\n",
-					dv->devname, strerror(errno));
+				fprintf(stderr, Name ": hot remove failed "
+					"for %s: %s\n",	dnprintable,
+					strerror(errno));
 				return 1;
 			}
 			if (verbose >= 0)
-				fprintf(stderr, Name ": hot removed %s\n", dv->devname);
+				fprintf(stderr, Name ": hot removed %s\n",
+					dnprintable);
 			break;
 
 		case 'f': /* set faulty */
 			/* FIXME check current member */
 			if (ioctl(fd, SET_DISK_FAULTY, (unsigned long) stb.st_rdev)) {
 				fprintf(stderr, Name ": set device faulty failed for %s:  %s\n",
-					dv->devname, strerror(errno));
+					dnprintable, strerror(errno));
 				return 1;
 			}
 			if (verbose >= 0)
 				fprintf(stderr, Name ": set %s faulty in %s\n",
-					dv->devname, devname);
+					dnprintable, devname);
 			break;
 		}
 	}
 	return 0;
 	
 }
+
+int autodetect(void)
+{
+	/* Open any md device, and issue the RAID_AUTORUN ioctl */
+	int rv = 1;
+	int fd = dev_open("9:0", O_RDONLY);
+	if (fd >= 0) {
+		if (ioctl(fd, RAID_AUTORUN, 0) == 0)
+			rv = 0;
+		close(fd);
+	}
+	return rv;
+}
+#endif
