@@ -114,6 +114,8 @@ struct mdstat_ent *mdstat_read(int hold, int start)
 		f = fopen("/proc/mdstat", "r");
 	if (f == NULL)
 		return NULL;
+	else
+		fcntl(fileno(f), F_SETFD, FD_CLOEXEC);
 
 	all = NULL;
 	end = &all;
@@ -221,8 +223,10 @@ struct mdstat_ent *mdstat_read(int hold, int start)
 			end = &ent->next;
 		}
 	}
-	if (hold && mdstat_fd == -1)
+	if (hold && mdstat_fd == -1) {
 		mdstat_fd = dup(fileno(f));
+		fcntl(mdstat_fd, F_SETFD, FD_CLOEXEC);
+	}
 	fclose(f);
 
 	/* If we might want to start array,
@@ -251,3 +255,15 @@ void mdstat_wait(int seconds)
 	tm.tv_usec = 0;
 	select(mdstat_fd >2 ? mdstat_fd+1:3, NULL, NULL, &fds, &tm);
 }
+
+int mddev_busy(int devnum)
+{
+	struct mdstat_ent *mdstat = mdstat_read(0, 0);
+	struct mdstat_ent *me;
+
+	for (me = mdstat ; me ; me = me->next)
+		if (me->devnum == devnum)
+			break;
+	free_mdstat(mdstat);
+	return me != NULL;
+}
